#include "tga_rw.h"



static void DecompressTGA(Byte*, TGAHEAD*);
static Byte* CompressTGA(Byte*, TGAHEAD*, int*);
static void WriteSameData(Byte*, Byte*, Byte*, int*);
static void WriteDifferentData(Byte*, Byte*, Byte*, int*);

Byte offset=0x80;






bool ReadTGA(char *filename, Byte **pixel_addr, TGAHEAD *head)
{
	int i, data_num=0, src_num;
	Byte *pixel_data;
	FILE *read;

	memset(head, 0, sizeof(TGAHEAD));

	read=fopen(filename, "rb");
	if (read==NULL) {
		printf("%s not found\n", filename);
		return FALSE;
	}

	fread(&head->No_ID, sizeof(Byte), 1, read);
	fread(&head->CMap_Type, sizeof(Byte), 1, read);
	fread(&head->image_type, sizeof(Byte), 1, read);
	fread(&head->CMap_Spec, sizeof(Byte), 5, read);
	fread(&head->Ox, sizeof(Word), 1, read);
	fread(&head->Oy, sizeof(Word), 1, read);
	fread(&head->width, sizeof(Word), 1, read);
	fread(&head->height, sizeof(Word), 1, read);
	fread(&head->pixel, sizeof(Byte), 1, read);
	fread(&head->IDesc_Type, sizeof(Byte), 1, read);

	if (head->image_type!=2 && head->image_type!=10) {
		printf("%s is not tga file\n", filename);
		return FALSE;
	}
	else if(head->pixel!=32) {
		printf("%s is not 32bit tga file\n", filename);
		return FALSE;
	}

	src_num=head->width*head->height*4;
	pixel_data=new Byte[src_num];

	for(i=0; i<src_num; i++){
		if( fread(&pixel_data[i], sizeof(Byte), 1, read)!=0 )
			data_num++;
		else
			break;
	}

	if (head->image_type==10)
		DecompressTGA(pixel_data, head);

	*pixel_addr=pixel_data;

	return TRUE;
}


static void DecompressTGA(Byte *cp_pixel, TGAHEAD *head)
{
	int i, j;
	int index=0, data_num=0, src_num=head->width*head->height*4;
	Byte dist, *dc_pixel;
	dc_pixel=new Byte[src_num];


	while( data_num<src_num ) {
		dist=cp_pixel[index];
		index++;

		if( dist<offset) {
			for(i=0; i<dist+1; i++) {
				for(j=0; j<4; j++)
					dc_pixel[data_num+j]=cp_pixel[index+j];

				data_num+=4;
				index+=4;
			}
		}
		else {
			for(i=offset; i<dist+1; i++) {
				for(j=0; j<4; j++)
					dc_pixel[data_num+j]=cp_pixel[index+j];

				data_num+=4;
			}
			index+=4;
		}
	}

	for (i=0; i<src_num; i++)
		cp_pixel[i]=dc_pixel[i];
	delete[] dc_pixel;

	head->image_type=2;
}




bool WriteTGA(char *filename, Byte *pixel, TGAHEAD *head)
{
	int data_num;
	Byte *comp_pixel;
	FILE *write;

	if(head->image_type!=2 || head->pixel!=32) {
		return FALSE;
	}

	comp_pixel=CompressTGA(pixel, head, &data_num);

	write=fopen(filename, "wb");
	fwrite(&head->No_ID, sizeof(Byte), 1, write);
	fwrite(&head->CMap_Type, sizeof(Byte), 1, write);
	fwrite(&head->image_type, sizeof(Byte), 1, write);
	fwrite(&head->CMap_Spec, sizeof(Byte), 5, write);
	fwrite(&head->Ox, sizeof(Word), 1, write);
	fwrite(&head->Oy, sizeof(Word), 1, write);
	fwrite(&head->width, sizeof(Word), 1, write);
	fwrite(&head->height, sizeof(Word), 1, write);
	fwrite(&head->pixel, sizeof(Byte), 1, write);
	fwrite(&head->IDesc_Type, sizeof(Byte), 1, write);

	for(int i=0; i<data_num; i++)
		fwrite( &comp_pixel[i], sizeof(Byte), 1, write);
	fclose(write);

	if (comp_pixel!=pixel) {
		delete[] comp_pixel;
		head->image_type=2;
	}


	return TRUE;
}







static Byte* CompressTGA(Byte *pixel_data, TGAHEAD *head, int *data_num)
{
	int x,y,j;
	int width=head->width, height=head->height;
	Byte dist_pixel[4], same=offset, diff=0;
	Byte *temp_pixel, *comp_pixel;

	comp_pixel=new Byte[width*height*5];
	temp_pixel=new Byte[width*4];
	*data_num=0;

	for(y=0; y<height; y++){
		for(x=0; x<width; x++){
			for (j=0; j<4; j++)
				dist_pixel[j]=pixel_data[(y*width+x)*4+j];

			if (	dist_pixel[0]==pixel_data[(y*width+x+1)*4+0]
				&& dist_pixel[1]==pixel_data[(y*width+x+1)*4+1]
				&& dist_pixel[2]==pixel_data[(y*width+x+1)*4+2]
				&& dist_pixel[3]==pixel_data[(y*width+x+1)*4+3]
				&& x%width!=width-1
				&& same<0xff ) {
				same++;

				if(diff!=0)
					WriteDifferentData(&diff, temp_pixel, comp_pixel, data_num);
			}
			else  {
				if(same!=offset) {
					WriteSameData(&same, dist_pixel, comp_pixel, data_num);
				}
				else {
					for (j=0; j<4; j++) {
						temp_pixel[diff*4+j]=dist_pixel[j];
					}
					diff++;

					if(diff==0x79 ) {
						WriteDifferentData(&diff, temp_pixel, comp_pixel, data_num);
					}
				}
			}

		}

		if(same!=offset)
			WriteSameData(&same, dist_pixel, comp_pixel, data_num);
		else if(diff!=0)
			WriteDifferentData(&diff, temp_pixel, comp_pixel, data_num);
	}

	if(*data_num<width*height*4) {
		head->image_type=10;
		return comp_pixel;
	}
	else {
		delete comp_pixel;
		return pixel_data;
	}
}

static void WriteSameData(Byte *same, Byte *temp_pixel, Byte *comp_pixel, int *data_num)
{
	comp_pixel[*data_num]=*same;
	*data_num+=1;

	for (int j=0; j<4; j++)
		comp_pixel[*data_num+j]=temp_pixel[j];

	*data_num+=4;
	*same=offset;
}

static void WriteDifferentData(Byte *diff, Byte *temp_pixel, Byte *comp_pixel, int *data_num)
{
	comp_pixel[*data_num]=*diff-1;
	*data_num+=1;

	for (int j=0; j<*diff*4; j++)
		comp_pixel[*data_num+j]=temp_pixel[j];

	*data_num+=4*(*diff);
	*diff=0;
}



